﻿using System;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using JsonRpcLite.Rpc;
using KcpNet;

namespace JsonRpcLite.Kcp
{
    public class JsonRpcKcpClientEngine:IJsonRpcClientEngine
    {
        private readonly KcpClient _client;

        public string Name { get; }


        public JsonRpcKcpClientEngine(IPAddress address, int port, int connectTimeout = Timeout.Infinite)
        {
            Name = nameof(JsonRpcKcpClientEngine);
            var remoteEndPoint = new IPEndPoint(address, port);
            _client = new KcpClient();
            _client.Connect(remoteEndPoint, connectTimeout);
        }


        private async Task<byte[]> ReadBytesAsync(int size, CancellationToken cancellationToken)
        {
            if (size > 0)
            {
                var bytes = new byte[size];
                var length = size;
                var startIndex = 0;
                while (length > 0)
                {
                    var buffer = new byte[length];
                    var bytesRead = await _client.ReadAsync(buffer, cancellationToken).ConfigureAwait(false);
                    if (bytesRead == 0)
                    {
                        throw new InvalidOperationException("Connection closed.");
                    }
                    if (bytesRead > 0)
                    {
                        Array.Copy(buffer, 0, bytes, startIndex, bytesRead);
                        startIndex += bytesRead;
                        length -= bytesRead;
                    }
                }

                return bytes;
            }

            return null;
        }

        private async Task SendStringAsync(string str, CancellationToken cancellationToken)
        {
            var strData = Encoding.UTF8.GetBytes(str);
            var lengthData = BitConverter.GetBytes(strData.Length);
            await _client.SendAsync(lengthData, cancellationToken).ConfigureAwait(false);
            await _client.SendAsync(strData, cancellationToken).ConfigureAwait(false);
        }

        public async Task<string> ProcessAsync(string serviceName, string requestString, CancellationToken cancellationToken)
        {
            var requestData = Encoding.UTF8.GetBytes(requestString);
            var responseData = await ProcessAsync(serviceName, requestData, cancellationToken).ConfigureAwait(false);
            return Encoding.UTF8.GetString(responseData);
        }

        public async Task<byte[]> ProcessAsync(string serviceName, byte[] requestData,CancellationToken cancellationToken)
        {
            await SendStringAsync(serviceName,cancellationToken).ConfigureAwait(false);
            
            var lengthData = BitConverter.GetBytes(requestData.Length);
            await _client.SendAsync(lengthData ,cancellationToken).ConfigureAwait(false);
            await _client.SendAsync(requestData, cancellationToken).ConfigureAwait(false);

            lengthData = await ReadBytesAsync(4, cancellationToken).ConfigureAwait(false);
            var length = BitConverter.ToInt32(lengthData);
            var responseData = await ReadBytesAsync(length, cancellationToken).ConfigureAwait(false);
            return responseData;
        }

        public async Task CloseAsync()
        {
            _client.Close();
            await Task.CompletedTask;
        }


    }
}
